import os
import time
from typing import Any, Dict, List, Tuple, Type

import httpx

from exchange import Message, Tool
from exchange.content import Text, ToolResult, ToolUse
from exchange.providers.base import Provider, Usage
from exchange.providers.utils import raise_for_status

ANTHROPIC_HOST = "https://api.anthropic.com/v1/messages"


class AnthropicProvider(Provider):
    def __init__(self, client: httpx.Client) -> None:
        self.client = client

    @classmethod
    def from_env(cls: Type["AnthropicProvider"]) -> "AnthropicProvider":
        url = os.environ.get("ANTHROPIC_HOST", ANTHROPIC_HOST)
        try:
            key = os.environ["ANTHROPIC_API_KEY"]
        except KeyError:
            raise RuntimeError("Failed to get ANTHROPIC_API_KEY from the environment")
        client = httpx.Client(
            base_url=url,
            headers={
                "x-api-key": key,
                "content-type": "application/json",
                "anthropic-version": "2023-06-01",
            },
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)

    @staticmethod
    def get_usage(data: Dict) -> Usage:  # noqa: ANN401
        usage = data.get("usage")
        input_tokens = usage.get("input_tokens")
        output_tokens = usage.get("output_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    @staticmethod
    def anthropic_response_to_message(response: Dict) -> Message:
        content_blocks = response.get("content", [])
        content = []
        for block in content_blocks:
            if block["type"] == "text":
                content.append(Text(text=block["text"]))
            elif block["type"] == "tool_use":
                content.append(
                    ToolUse(
                        id=block["id"],
                        name=block["name"],
                        parameters=block["input"],
                    )
                )
        return Message(role="assistant", content=content)

    @staticmethod
    def tools_to_anthropic_spec(tools: Tuple[Tool]) -> List[Dict[str, Any]]:
        return [
            {
                "name": tool.name,
                "description": tool.description or "",
                "input_schema": tool.parameters,
            }
            for tool in tools
        ]

    @staticmethod
    def messages_to_anthropic_spec(messages: List[Message]) -> List[Dict[str, Any]]:
        messages_spec = []
        # if messages is empty - just make a default
        for message in messages:
            converted = {"role": message.role}
            for content in message.content:
                if isinstance(content, Text):
                    converted["content"] = [{"type": "text", "text": content.text}]
                elif isinstance(content, ToolUse):
                    converted.setdefault("content", []).append(
                        {
                            "type": "tool_use",
                            "id": content.id,
                            "name": content.name,
                            "input": content.parameters,
                        }
                    )
                elif isinstance(content, ToolResult):
                    converted.setdefault("content", []).append(
                        {
                            "type": "tool_result",
                            "tool_use_id": content.tool_use_id,
                            "content": content.output,
                        }
                    )
            messages_spec.append(converted)
        if len(messages_spec) == 0:
            converted = {
                "role": "user",
                "content": [{"type": "text", "text": "Ignore"}],
            }
            messages_spec.append(converted)
        return messages_spec

    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: List[Tool] = [],
        **kwargs: Dict[str, Any],
    ) -> Tuple[Message, Usage]:
        tools_set = set()
        unique_tools = []
        for tool in tools:
            if tool.name not in tools_set:
                unique_tools.append(tool)
                tools_set.add(tool.name)

        payload = dict(
            system=system,
            model=model,
            max_tokens=4096,
            messages=self.messages_to_anthropic_spec(messages),
            tools=self.tools_to_anthropic_spec(tuple(unique_tools)),
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}

        max_retries = 5
        initial_wait = 10  # Start with 10 seconds
        backoff_factor = 1
        for retry in range(max_retries):
            response = self.client.post(ANTHROPIC_HOST, json=payload)
            if response.status_code not in (429, 529, 500):
                break
            else:
                sleep_time = initial_wait + (backoff_factor * (2**retry))
                time.sleep(sleep_time)

        if response.status_code in (429, 529, 500):
            raise httpx.HTTPStatusError(
                f"Failed after {max_retries} retries due to rate limiting",
                request=response.request,
                response=response,
            )

        response_data = raise_for_status(response).json()
        message = self.anthropic_response_to_message(response_data)
        usage = self.get_usage(response_data)

        return message, usage



from typing import Literal, Optional, Type

from exchange.moderators import PassiveModerator
from exchange.moderators.base import Moderator

MAX_TOKENS = 112000


def pop_checkpoint(
    exchange: Type["exchange.exchange.Exchange"],  # noqa: F821
    exclude_last: Literal[0, 1] = 0,
) -> Type["exchange.exchange.Exchange"]:  # noqa: F821
    """Pop messages from the front of the list in sections"""
    checkpoint = exchange.checkpoints.pop(0)
    for _ in range(checkpoint.end_index - checkpoint.start_index - exclude_last):
        exchange.messages.pop(checkpoint.start_index)

    # rewrite checkpoint indexes
    for cp in exchange.checkpoints:
        cp.start_index = cp.start_index - checkpoint.end_index + exclude_last
        cp.end_index = cp.end_index - checkpoint.end_index + exclude_last

    if exclude_last:
        # there is still one entry in the first checkpoint only
        checkpoint.token_count = checkpoint.latest_generated_tokens
        checkpoint.end_index = exclude_last
        exchange.checkpoints.insert(0, checkpoint)

    return exchange


class ContextTruncate(Moderator):
    def __init__(self, model: Optional[str] = None) -> None:
        self.model = model
        self.system_prompt_token_count = None

    def rewrite(self, exchange: Type["exchange.exchange.Exchange"]) -> None:  # noqa: F821
        """Truncate the exchange messages with a FIFO strategy."""
        if not self.system_prompt_token_count:
            # calculate the system prompt tokens (includes functions etc...)
            _system_token_exchange = exchange.replace(
                messages=[],
                checkpoints=[],
                moderator=PassiveModerator(),
                model=self.model if self.model else exchange.model,
            )
            _ = _system_token_exchange.generate()
            checkpoint = _system_token_exchange.checkpoints.pop()
            self.system_prompt_token_count = checkpoint.token_count - checkpoint.latest_generated_tokens

        while sum(cp.token_count for cp in exchange.checkpoints) > MAX_TOKENS:
            exchange = pop_checkpoint(exchange)

            # if first message is a tool_result, it means there's no matching pair. remove it
            if exchange.messages[0].tool_result:
                pop_checkpoint(exchange, exclude_last=1)

            # Update the token count on the the first checkpoint to reflect the system prompt
            exchange.checkpoints[0].token_count += self.system_prompt_token_count



import os
from unittest.mock import patch

import pytest
from exchange import Message, Text
from exchange.providers.databricks import DatabricksProvider


@pytest.fixture
@patch.dict(
    os.environ,
    {"DATABRICKS_HOST": "http://test-host", "DATABRICKS_TOKEN": "test_token"},
)
def databricks_provider():
    return DatabricksProvider.from_env()


@patch("httpx.Client.post")
@patch("time.sleep", return_value=None)
@patch("logging.warning")
@patch("logging.error")
def test_databricks_completion(mock_error, mock_warning, mock_sleep, mock_post, databricks_provider):
    mock_response = {
        "choices": [{"message": {"role": "assistant", "content": "Hello!"}}],
        "usage": {"prompt_tokens": 10, "completion_tokens": 25, "total_tokens": 35},
    }
    mock_post.return_value.json.return_value = mock_response

    model = "my-databricks-model"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    tools = ()

    reply_message, reply_usage = databricks_provider.complete(
        model=model, system=system, messages=messages, tools=tools
    )

    assert reply_message.content == [Text(text="Hello!")]
    assert reply_usage.total_tokens == 35
    assert mock_post.call_count == 1
    mock_post.assert_called_once_with(
        "serving-endpoints/my-databricks-model/invocations",
        json={
            "messages": [
                {"role": "system", "content": system},
                {"role": "user", "content": "Hello"},
            ]
        },
    )



import json
import traceback
from copy import deepcopy
from typing import Any, Dict, List, Mapping, Tuple

from attrs import define, evolve, field
from tiktoken import get_encoding

from exchange.checkpoint import Checkpoint
from exchange.content import ToolResult, ToolUse
from exchange.message import Message
from exchange.moderators import ContextTruncate, Moderator
from exchange.providers import Provider, Usage
from exchange.tool import Tool


def validate_tool_output(output: str) -> None:
    """Validate tool output for the given model"""
    max_output_chars = 2**20
    max_output_tokens = 16000
    encoder = get_encoding("cl100k_base")
    if len(output) > max_output_chars or len(encoder.encode(output)) > max_output_tokens:
        raise ValueError("This tool call created an output that was too long to handle!")


@define(frozen=True)
class Exchange:
    """An exchange of messages with an LLM

    The exchange class is meant to be largely immutable, with only the message list
    growing once constructed. Use .replace to alter the model, tools, etc.

    The exchange supports tool usage, calling tools and letting the model respond when
    using the .reply method. It handles most forms of errors and sends those errors back
    to the model, to let it attempt to recover.
    """

    provider: Provider
    model: str
    system: str
    moderator: Moderator = field(default=ContextTruncate())
    tools: Tuple[Tool] = field(factory=tuple, converter=tuple)
    messages: List[Message] = field(factory=list)
    checkpoints: List[Checkpoint] = field(factory=list)

    @property
    def _toolmap(self) -> Mapping[str, Tool]:
        return {tool.name: tool for tool in self.tools}

    def replace(self, **kwargs: Dict[str, Any]) -> "Exchange":
        """Make a copy of the exchange, replacing any passed arguments"""
        if kwargs.get("messages") is None:
            kwargs["messages"] = deepcopy(self.messages)
        if kwargs.get("checkpoints") is None:
            kwargs["checkpoints"] = deepcopy(self.checkpoints)
        return evolve(self, **kwargs)

    def add(self, message: Message) -> None:
        """Add a message to the history."""
        if self.messages and message.role == self.messages[-1].role:
            raise ValueError("Messages in the exchange must alternate between user and assistant")
        self.messages.append(message)

    def generate(self) -> Message:
        """Generate the next message."""
        self.moderator.rewrite(self)

        message, usage = self.provider.complete(
            self.model,
            self.system,
            messages=self.messages,
            tools=self.tools,
        )

        self.add(message)
        # this has to come after adding the response
        self.add_checkpoint(usage)

        return message

    def reply(self, max_tool_use: int = 128) -> Message:
        """Get the reply from the underlying model.

        This will process any requests for tool calls, calling them immediately, and
        storing the intermediate tool messages in the queue. It will return after the
        first response that does not request a tool use

        Args:
            max_tool_use: The maximum number of tool calls to make before returning. Defaults to 128.
        """
        if max_tool_use <= 0:
            raise ValueError("max_tool_use must be greater than 0")
        response = self.generate()
        curr_iter = 1  # generate() already called once
        while response.tool_use:
            content = []
            for tool_use in response.tool_use:
                tool_result = self.call_function(tool_use)
                content.append(tool_result)
            self.add(Message(role="user", content=content))

            # We've reached the limit of tool calls - break out of the loop
            if curr_iter >= max_tool_use:
                # At this point, the most recent message is `Message(role='user', content=ToolResult(...))`
                response = Message.assistant(
                    f"We've stopped executing additional tool cause because we reached the limit of {max_tool_use}",
                )
                self.add(response)
                break
            else:
                response = self.generate()
                curr_iter += 1

        return response

    def call_function(self, tool_use: ToolUse) -> ToolResult:
        """Call the function indicated by the tool use"""
        tool = self._toolmap.get(tool_use.name)

        if tool is None or tool_use.is_error:
            output = f"ERROR: Failed to use tool {tool_use.id}.\nDo NOT use the same tool name and parameters again - that will lead to the same error."  # noqa: E501

            if tool_use.is_error:
                output += f"\n{tool_use.error_message}"
            elif tool is None:
                valid_tool_names = ", ".join(self._toolmap.keys())
                output += f"\nNo tool exists with the name '{tool_use.name}'. Valid tool names are: {valid_tool_names}"

            return ToolResult(tool_use_id=tool_use.id, output=output, is_error=True)

        try:
            if isinstance(tool_use.parameters, dict):
                output = json.dumps(tool.function(**tool_use.parameters))
            elif isinstance(tool_use.parameters, list):
                output = json.dumps(tool.function(*tool_use.parameters))
            else:
                raise ValueError(
                    f"The provided tool parameters, {tool_use.parameters} could not be interpreted as a mapping of arguments."  # noqa: E501
                )

            validate_tool_output(output)

            is_error = False
        except Exception as e:
            tb = traceback.format_exc()
            output = str(tb) + "\n" + str(e)
            is_error = True

        return ToolResult(tool_use_id=tool_use.id, output=output, is_error=is_error)

    def add_tool_use(self, tool_use: ToolUse) -> None:
        """Manually add a tool use and corresponding result

        This will call the implied function and add an assistant
        message requesting the ToolUse and a user message with the ToolResult
        """
        tool_result = self.call_function(tool_use)
        self.add(Message(role="assistant", content=[tool_use]))
        self.add(Message(role="user", content=[tool_result]))

    def add_checkpoint(self, usage: Usage) -> None:
        self.checkpoints.append(
            Checkpoint(
                start_index=(0 if not self.checkpoints else self.checkpoints[-1].end_index),
                end_index=len(self.messages),
                token_count=(
                    usage.total_tokens
                    if not self.checkpoints
                    else usage.total_tokens - sum(cp.token_count for cp in self.checkpoints)
                ),
                latest_generated_tokens=usage.output_tokens,
            )
        )



import os
from typing import Any, Dict, List, Tuple, Type

import httpx

from exchange.message import Message
from exchange.providers.base import Provider, Usage
from exchange.providers.utils import (
    messages_to_openai_spec,
    openai_response_to_message,
    raise_for_status,
    tools_to_openai_spec,
)
from exchange.tool import Tool


class DatabricksProvider(Provider):
    """Provides chat completions for models on Databricks serving endpoints

    Models are expected to follow the llm/v1/chat "task". This includes support
    for foundation and external model endpoints
    https://docs.databricks.com/en/machine-learning/model-serving/create-foundation-model-endpoints.html#create-generative-ai-model-serving-endpoints
    """

    def __init__(self, client: httpx.Client) -> None:
        super().__init__()
        self.client = client

    @classmethod
    def from_env(cls: Type["DatabricksProvider"]) -> "DatabricksProvider":
        try:
            url = os.environ["DATABRICKS_HOST"]
        except KeyError:
            raise RuntimeError(
                "Failed to get DATABRICKS_HOST from the environment. See https://docs.databricks.com/en/dev-tools/auth/index.html#general-host-token-and-account-id-environment-variables-and-fields"
            )
        try:
            key = os.environ["DATABRICKS_TOKEN"]
        except KeyError:
            raise RuntimeError(
                "Failed to get DATABRICKS_TOKEN from the environment. See https://docs.databricks.com/en/dev-tools/auth/index.html#general-host-token-and-account-id-environment-variables-and-fields"
            )
        client = httpx.Client(
            base_url=url,
            auth=("token", key),
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.pop("usage")
        input_tokens = usage.get("prompt_tokens")
        output_tokens = usage.get("completion_tokens")
        total_tokens = usage.get("total_tokens")
        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: Tuple[Tool],
        **kwargs: Dict[str, Any],
    ) -> Tuple[Message, Usage]:
        payload = dict(
            messages=[
                {"role": "system", "content": system},
                *messages_to_openai_spec(messages),
            ],
            tools=tools_to_openai_spec(tools) if tools else [],
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}
        response = self.client.post(
            f"serving-endpoints/{model}/invocations",
            json=payload,
        )
        data = raise_for_status(response).json()
        message = openai_response_to_message(data)
        usage = self.get_usage(data)
        return message, usage



from functools import cache
from typing import Type

from exchange.providers.anthropic import AnthropicProvider  # noqa
from exchange.providers.base import Provider, Usage  # noqa
from exchange.providers.databricks import DatabricksProvider  # noqa
from exchange.providers.openai import OpenAiProvider  # noqa
from exchange.utils import load_plugins


@cache
def get_provider(name: str) -> Type[Provider]:
    return load_plugins(group="exchange.provider")[name]



import logging
import os
from unittest.mock import patch

import pytest
from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.providers.bedrock import BedrockProvider
from exchange.tool import Tool

logger = logging.getLogger(__name__)


@pytest.fixture
@patch.dict(
    os.environ,
    {
        "AWS_REGION": "us-east-1",
        "AWS_ACCESS_KEY_ID": "fake-access-key",
        "AWS_SECRET_ACCESS_KEY": "fake-secret-key",
        "AWS_SESSION_TOKEN": "fake-session-token",
    },
)
def bedrock_provider():
    return BedrockProvider.from_env()


@patch("time.time", return_value=1624250000)
def test_sign_and_get_headers(mock_time, bedrock_provider):
    # Create sample values
    method = "POST"
    url = "https://bedrock-runtime.us-east-1.amazonaws.com/some/path"
    payload = {"key": "value"}
    service = "bedrock"
    # Generate headers
    headers = bedrock_provider.client.sign_and_get_headers(
        method,
        url,
        payload,
        service,
    )
    # Assert that headers contain expected keys
    assert "Authorization" in headers
    assert "Content-Type" in headers
    assert "X-Amz-date" in headers
    assert "x-amz-content-sha256" in headers
    assert "X-Amz-Security-Token" in headers


@patch("httpx.Client.post")
def test_complete(mock_post, bedrock_provider):
    # Mocked response from the server
    mock_response = {
        "output": {"message": {"role": "assistant", "content": [{"text": "Hello, world!"}]}},
        "usage": {"inputTokens": 10, "outputTokens": 15, "totalTokens": 25},
    }
    mock_post.return_value.json.return_value = mock_response

    model = "test-model"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    tools = ()

    reply_message, reply_usage = bedrock_provider.complete(model=model, system=system, messages=messages, tools=tools)

    # Assertions for reply message
    assert reply_message.content[0].text == "Hello, world!"
    assert reply_usage.total_tokens == 25


def test_message_to_bedrock_spec_text(bedrock_provider):
    message = Message(role="user", content=[Text("Hello, world!")])
    expected = {"role": "user", "content": [{"text": "Hello, world!"}]}
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_use(bedrock_provider):
    tool_use = ToolUse(id="tool-1", name="WordCount", parameters={"text": "Hello, world!"})
    message = Message(role="assistant", content=[tool_use])
    expected = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_result(bedrock_provider):
    message = Message(
        role="assistant",
        content=[ToolUse(id="tool-1", name="WordCount", parameters={"text": "Hello, world!"})],
    )
    expected = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_tool_result_text(bedrock_provider):
    tool_result = ToolResult(tool_use_id="tool-1", output="Error occured", is_error=True)
    message = Message(role="user", content=[tool_result])
    expected = {
        "role": "user",
        "content": [
            {
                "toolResult": {
                    "toolUseId": "tool-1",
                    "content": [{"text": "Error occured"}],
                    "status": "error",
                }
            }
        ],
    }
    assert bedrock_provider.message_to_bedrock_spec(message) == expected


def test_message_to_bedrock_spec_invalid(bedrock_provider):
    with pytest.raises(Exception):
        bedrock_provider.message_to_bedrock_spec(Message(role="user", content=[]))


def test_response_to_message_text(bedrock_provider):
    response = {"role": "user", "content": [{"text": "Hello, world!"}]}
    message = bedrock_provider.response_to_message(response)
    assert message.role == "user"
    assert message.content[0].text == "Hello, world!"


def test_response_to_message_tool_use(bedrock_provider):
    response = {
        "role": "assistant",
        "content": [
            {
                "toolUse": {
                    "toolUseId": "tool-1",
                    "name": "WordCount",
                    "input": {"text": "Hello, world!"},
                }
            }
        ],
    }
    message = bedrock_provider.response_to_message(response)
    assert message.role == "assistant"
    assert message.content[0].name == "WordCount"
    assert message.content[0].parameters == {"text": "Hello, world!"}


def test_response_to_message_tool_result(bedrock_provider):
    response = {
        "role": "user",
        "content": [
            {
                "toolResult": {
                    "toolResultId": "tool-1",
                    "content": [{"json": {"result": 2}}],
                }
            }
        ],
    }
    message = bedrock_provider.response_to_message(response)
    assert message.role == "user"
    assert message.content[0].tool_use_id == "tool-1"
    assert message.content[0].output == {"result": 2}


def test_response_to_message_invalid(bedrock_provider):
    with pytest.raises(Exception):
        bedrock_provider.response_to_message({})


def test_tools_to_bedrock_spec(bedrock_provider):
    def word_count(text: str):
        return len(text.split())

    tool = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    expected = {
        "tools": [
            {
                "toolSpec": {
                    "name": "WordCount",
                    "description": "Counts words.",
                    "inputSchema": {"json": {"text": "string"}},
                }
            }
        ]
    }
    assert bedrock_provider.tools_to_bedrock_spec((tool,)) == expected


def test_tools_to_bedrock_spec_duplicate(bedrock_provider):
    def word_count(text: str):
        return len(text.split())

    tool = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    tool_duplicate = Tool(
        name="WordCount",
        description="Counts words.",
        parameters={"text": "string"},
        function=word_count,
    )
    tools = bedrock_provider.tools_to_bedrock_spec((tool, tool_duplicate))
    assert set(tool["toolSpec"]["name"] for tool in tools["tools"]) == {"WordCount"}



"""Classes for interacting with the exchange API."""

from exchange.tool import Tool  # noqa
from exchange.content import Text, ToolResult, ToolUse  # noqa
from exchange.message import Message  # noqa
from exchange.exchange import Exchange  # noqa



from abc import ABC, abstractmethod
from typing import Type


class Moderator(ABC):
    @abstractmethod
    def rewrite(self, exchange: Type["exchange.exchange.Exchange"]) -> None:  # noqa: F821
        pass



import base64
import json
import re
from typing import Any, Dict, List, Tuple

import httpx
from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.tool import Tool


def raise_for_status(response: httpx.Response) -> httpx.Response:
    """Raise with reason text."""
    try:
        response.raise_for_status()
        return response
    except httpx.HTTPStatusError as e:
        response.read()
        if response.text:
            raise httpx.HTTPStatusError(f"{e}\n{response.text}", request=e.request, response=e.response)
        else:
            raise e


def encode_image(image_path: str) -> str:
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def messages_to_openai_spec(messages: List[Message]) -> List[Dict[str, Any]]:
    messages_spec = []
    for message in messages:
        converted = {"role": message.role}
        output = []
        for content in message.content:
            if isinstance(content, Text):
                converted["content"] = content.text
            elif isinstance(content, ToolUse):
                sanitized_name = re.sub(r"[^a-zA-Z0-9_-]", "_", content.name)
                converted.setdefault("tool_calls", []).append(
                    {
                        "id": content.id,
                        "type": "function",
                        "function": {
                            "name": sanitized_name,
                            "arguments": json.dumps(content.parameters),
                        },
                    }
                )
            elif isinstance(content, ToolResult):
                if content.output.startswith('"image:'):
                    image_path = content.output.replace('"image:', "").replace('"', "")
                    output.append(
                        {
                            "role": "tool",
                            "content": [
                                {
                                    "type": "text",
                                    "text": "This tool result included an image that is uploaded in the next message.",
                                },
                            ],
                            "tool_call_id": content.tool_use_id,
                        }
                    )
                    # Note: it is possible to only do this when message == messages[-1]
                    # but it doesn't seem to hurt too much with tokens to keep this.
                    output.append(
                        {
                            "role": "user",
                            "content": [
                                {
                                    "type": "image_url",
                                    "image_url": {"url": f"data:image/jpeg;base64,{encode_image(image_path)}"},
                                }
                            ],
                        }
                    )

                else:
                    output.append(
                        {
                            "role": "tool",
                            "content": content.output,
                            "tool_call_id": content.tool_use_id,
                        }
                    )

        if "content" in converted or "tool_calls" in converted:
            output = [converted] + output
        messages_spec.extend(output)
    return messages_spec


def tools_to_openai_spec(tools: Tuple[Tool]) -> Dict[str, Any]:
    tools_names = set()
    result = []
    for tool in tools:
        if tool.name in tools_names:
            # we should never allow duplicate tools
            raise ValueError(f"Duplicate tool name: {tool.name}")
        result.append(
            {
                "type": "function",
                "function": {
                    "name": tool.name,
                    "description": tool.description,
                    "parameters": tool.parameters,
                },
            }
        )
        tools_names.add(tool.name)
    return result


def openai_response_to_message(response: dict) -> Message:
    original = response["choices"][0]["message"]
    content = []
    text = original.get("content")
    if text:
        content.append(Text(text=text))

    tool_calls = original.get("tool_calls")
    if tool_calls:
        for tool_call in tool_calls:
            try:
                function_name = tool_call["function"]["name"]
                # We occasionally see the model generate an invalid function name
                # sending this back to openai raises a validation error
                if not re.match(r"^[a-zA-Z0-9_-]+$", function_name):
                    content.append(
                        ToolUse(
                            id=tool_call["id"],
                            name=function_name,
                            parameters=tool_call["function"]["arguments"],
                            is_error=True,
                            error_message=f"The provided function name '{function_name}' had invalid characters, it must match this regex [a-zA-Z0-9_-]+",  # noqa: E501
                        )
                    )
                else:
                    content.append(
                        ToolUse(
                            id=tool_call["id"],
                            name=function_name,
                            parameters=json.loads(tool_call["function"]["arguments"]),
                        )
                    )
            except json.JSONDecodeError:
                content.append(
                    ToolUse(
                        id=tool_call["id"],
                        name=tool_call["function"]["name"],
                        parameters=tool_call["function"]["arguments"],
                        is_error=True,
                        error_message=f"Could not interpret tool use parameters for id {tool_call['id']}: {tool_call['function']['arguments']}",  # noqa: E501
                    )
                )

    return Message(role="assistant", content=content)


def openai_single_message_context_length_exceeded(error_dict: dict) -> None:
    code = error_dict.get("code")
    if code == "context_length_exceeded" or code == "string_above_max_length":
        raise InitialMessageTooLargeError(f"Input message too long. Message: {error_dict.get('message')}")


class InitialMessageTooLargeError(Exception):
    """Custom error raised when the first input message in an exchange is too large."""

    pass



import inspect
from typing import Any, Callable, Type

from attrs import define

from exchange.utils import json_schema, parse_docstring


@define
class Tool:
    """A tool that can be used by a model.

    Attributes:
        name (str): The name of the tool
        description (str): A description of what the tool does
        parameters dict[str, Any]: A json schema of the function signature
        function (Callable): The python function that powers the tool
    """

    name: str
    description: str
    parameters: dict[str, Any]
    function: Callable

    @classmethod
    def from_function(cls: Type["Tool"], func: Any) -> "Tool":  # noqa: ANN401
        """Create a tool instance from a function and its docstring

        The function must have a docstring - we require it to load the description
        and parameter descriptions. This also supports a class instance with a __call__
        method.
        """
        if inspect.isfunction(func) or inspect.ismethod(func):
            name = func.__name__
        else:
            name = func.__class__.__name__.lower()
            func = func.__call__

        description, param_descriptions = parse_docstring(func)
        schema = json_schema(func)

        # Set the 'description' field of the schema to the arg's docstring description
        for arg in param_descriptions:
            arg_name, arg_description = arg["name"], arg["description"]

            if arg_name not in schema["properties"]:
                raise ValueError(f"Argument {arg_name} found in docstring but not in schema")
            schema["properties"][arg_name]["description"] = arg_description

        return cls(
            name=name,
            description=description,
            parameters=schema,
            function=func,
        )



import attrs
from exchange.tool import Tool


def get_current_weather(location: str) -> None:
    """Get the current weather in a given location

    Args:
        location (str): The city and state, e.g. San Francisco, CA
    """
    pass


def test_load():
    tool = Tool.from_function(get_current_weather)

    expected = {
        "name": "get_current_weather",
        "description": "Get the current weather in a given location",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "The city and state, e.g. San Francisco, CA",
                },
            },
            "required": ["location"],
        },
        "function": get_current_weather,
    }

    assert attrs.asdict(tool) == expected


def another_function(
    param1: int,
    param2: str,
    param3: bool,
    param4: float,
    param5: list[int],
    param6: dict[str, float],
) -> None:
    """
    This is another example function with various types

    Args:
        param1 (int): Description for param1
        param2 (str): Description for param2
        param3 (bool): Description for param3
        param4 (float): Description for param4
        param5 (list[int]): Description for param5
        param6 (dict[str, float]): Description for param6
    """
    pass


def test_load_types():
    tool = Tool.from_function(another_function)
    expected_schema = {
        "type": "object",
        "properties": {
            "param1": {"type": "integer", "description": "Description for param1"},
            "param2": {"type": "string", "description": "Description for param2"},
            "param3": {"type": "boolean", "description": "Description for param3"},
            "param4": {"type": "number", "description": "Description for param4"},
            "param5": {
                "type": "array",
                "items": {"type": "integer"},
                "description": "Description for param5",
            },
            "param6": {
                "type": "object",
                "additionalProperties": {"type": "number"},
                "description": "Description for param6",
            },
        },
        "required": ["param1", "param2", "param3", "param4", "param5", "param6"],
    }
    assert tool.parameters == expected_schema


def numpy_function(param1: int, param2: str) -> None:
    """
    This function uses numpy style docstrings

    Parameters
    ----------
    param1 : int
        Description for param1
    param2 : str
        Description for param2
    """
    pass


def test_load_numpy_style():
    tool = Tool.from_function(numpy_function)
    expected_schema = {
        "type": "object",
        "properties": {
            "param1": {"type": "integer", "description": "Description for param1"},
            "param2": {"type": "string", "description": "Description for param2"},
        },
        "required": ["param1", "param2"],
    }
    assert tool.parameters == expected_schema


def sphinx_function(param1: int, param2: str, param3: bool) -> None:
    """
    This function uses sphinx style docstrings

    :param param1: Description for param1
    :type param1: int
    :param param2: Description for param2
    :type param2: str
    :param param3: Description for param3
    :type param3: bool
    """
    pass


def test_load_sphinx_style():
    tool = Tool.from_function(sphinx_function)
    expected_schema = {
        "type": "object",
        "properties": {
            "param1": {"type": "integer", "description": "Description for param1"},
            "param2": {"type": "string", "description": "Description for param2"},
            "param3": {"type": "boolean", "description": "Description for param3"},
        },
        "required": ["param1", "param2", "param3"],
    }
    assert tool.parameters == expected_schema


class FunctionLike:
    def __init__(self, state: int) -> None:
        self.state = state

    def __call__(self, param1: int) -> int:
        """Example

        Args:
            param1 (int): Description for param1
        """
        return self.state + param1


def test_load_stateful_class():
    tool = Tool.from_function(FunctionLike(1))
    expected_schema = {
        "type": "object",
        "properties": {
            "param1": {"type": "integer", "description": "Description for param1"},
        },
        "required": ["param1"],
    }
    assert tool.parameters == expected_schema
    assert tool.function(2) == 3



from typing import Any, Dict, Optional

from attrs import define, asdict


CONTENT_TYPES = {}


class Content:
    def __init_subclass__(cls, **kwargs: Dict[str, Any]) -> None:
        super().__init_subclass__(**kwargs)
        CONTENT_TYPES[cls.__name__] = cls

    def to_dict(self) -> Dict[str, Any]:
        data = asdict(self, recurse=True)
        data["type"] = self.__class__.__name__
        return data


@define
class Text(Content):
    text: str


@define
class ToolUse(Content):
    id: str
    name: str
    parameters: Any
    is_error: bool = False
    error_message: Optional[str] = None


@define
class ToolResult(Content):
    tool_use_id: str
    output: str
    is_error: bool = False



import pytest
from exchange import utils
from unittest.mock import patch
from exchange.message import Message
from exchange.content import Text, ToolResult
from exchange.providers.utils import messages_to_openai_spec, encode_image


def test_encode_image():
    image_path = "tests/test_image.png"
    encoded_image = encode_image(image_path)

    # Adjust this string based on the actual initial part of your base64-encoded image.
    expected_start = "iVBORw0KGgo"
    assert encoded_image.startswith(expected_start)


def test_create_object_id() -> None:
    prefix = "test"
    object_id = utils.create_object_id(prefix)
    assert object_id.startswith(prefix + "_")
    assert len(object_id) == len(prefix) + 1 + 24  # prefix + _ + 24 chars


def test_compact() -> None:
    content = "This   is \n\n   a test"
    compacted = utils.compact(content)
    assert compacted == "This is a test"


def test_parse_docstring() -> None:
    def dummy_func(a, b, c):
        """
        This function does something.

        Args:
            a (int): The first parameter.
            b (str): The second parameter.
            c (list): The third parameter.
        """
        pass

    description, parameters = utils.parse_docstring(dummy_func)
    assert description == "This function does something."
    assert parameters == [
        {"name": "a", "annotation": "int", "description": "The first parameter."},
        {"name": "b", "annotation": "str", "description": "The second parameter."},
        {"name": "c", "annotation": "list", "description": "The third parameter."},
    ]


def test_parse_docstring_no_description() -> None:
    def dummy_func(a, b, c):
        """
        Args:
            a (int): The first parameter.
            b (str): The second parameter.
            c (list): The third parameter.
        """
        pass

    with pytest.raises(ValueError) as e:
        utils.parse_docstring(dummy_func)

    assert "Attempted to load from a function" in str(e.value)


def test_json_schema() -> None:
    def dummy_func(a: int, b: str, c: list) -> None:
        pass

    schema = utils.json_schema(dummy_func)

    assert schema == {
        "type": "object",
        "properties": {
            "a": {"type": "integer"},
            "b": {"type": "string"},
            "c": {"type": "string"},
        },
        "required": ["a", "b", "c"],
    }


def test_load_plugins() -> None:
    class DummyEntryPoint:
        def __init__(self, name, plugin):
            self.name = name
            self.plugin = plugin

        def load(self):
            return self.plugin

    with patch("exchange.utils.entry_points") as entry_points_mock:
        entry_points_mock.return_value = [
            DummyEntryPoint("plugin1", object()),
            DummyEntryPoint("plugin2", object()),
        ]

        plugins = utils.load_plugins("dummy_group")

        assert "plugin1" in plugins
        assert "plugin2" in plugins
        assert len(plugins) == 2


def test_messages_to_openai_spec():
    # Use provided test image
    png_path = "tests/test_image.png"

    # Create a list of messages as input
    messages = [
        Message(role="user", content=[Text(text="Hello, Assistant!")]),
        Message(role="assistant", content=[Text(text="Here is a text with tool usage")]),
        Message(
            role="tool",
            content=[ToolResult(tool_use_id="1", output=f'"image:{png_path}')],
        ),
    ]

    # Call the function
    output = messages_to_openai_spec(messages)

    assert "This tool result included an image that is uploaded in the next message." in str(output)
    assert "{'role': 'user', 'content': [{'type': 'image_url'" in str(output)



from typing import Type
from exchange.moderators.base import Moderator


class PassiveModerator(Moderator):
    def rewrite(self, _: Type["exchange.exchange.Exchange"]) -> None:  # noqa: F821
        pass



from copy import deepcopy
import json
from unittest.mock import Mock
from attrs import asdict
import httpx
import pytest
from unittest.mock import patch

from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.providers.utils import (
    messages_to_openai_spec,
    openai_response_to_message,
    raise_for_status,
    tools_to_openai_spec,
)
from exchange.tool import Tool

OPEN_AI_TOOL_USE_RESPONSE = response = {
    "choices": [
        {
            "role": "assistant",
            "message": {
                "tool_calls": [
                    {
                        "id": "1",
                        "function": {
                            "name": "example_fn",
                            "arguments": json.dumps(
                                {
                                    "param": "value",
                                }
                            ),
                            # TODO: should this handle dict's as well?
                        },
                    }
                ],
            },
        }
    ],
    "usage": {
        "input_tokens": 10,
        "output_tokens": 25,
        "total_tokens": 35,
    },
}


def example_fn(param: str) -> None:
    """
    Testing function.

    Args:
        param (str): Description of param1
    """
    pass


def example_fn_two() -> str:
    """
    Second testing function

    Returns:
        str: Description of return value
    """
    pass


def test_raise_for_status_success() -> None:
    response = Mock(spec=httpx.Response)
    response.status_code = 200

    result = raise_for_status(response)

    assert result == response


def test_raise_for_status_failure_with_text() -> None:
    response = Mock(spec=httpx.Response)
    response.status_code = 404
    response.text = "Not Found: John Cena"

    try:
        raise_for_status(response)
    except httpx.HTTPStatusError as e:
        assert e.response == response
        assert str(e) == "404 Not Found: John Cena"
        assert e.request is None


def test_raise_for_status_failure_without_text() -> None:
    response = Mock(spec=httpx.Response)
    response.status_code = 500
    response.text = ""

    try:
        raise_for_status(response)
    except httpx.HTTPStatusError as e:
        assert e.response == response
        assert str(e) == "500 Internal Server Error"
        assert e.request is None


def test_messages_to_openai_spec() -> None:
    messages = [
        Message(role="assistant", content=[Text("Hello!")]),
        Message(role="user", content=[Text("How are you?")]),
        Message(
            role="assistant",
            content=[ToolUse(id=1, name="tool1", parameters={"param1": "value1"})],
        ),
        Message(role="user", content=[ToolResult(tool_use_id=1, output="Result")]),
    ]

    spec = messages_to_openai_spec(messages)

    assert spec == [
        {"role": "assistant", "content": "Hello!"},
        {"role": "user", "content": "How are you?"},
        {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": 1,
                    "type": "function",
                    "function": {
                        "name": "tool1",
                        "arguments": '{"param1": "value1"}',
                    },
                }
            ],
        },
        {
            "role": "tool",
            "content": "Result",
            "tool_call_id": 1,
        },
    ]


def test_tools_to_openai_spec() -> None:
    tools = (Tool.from_function(example_fn), Tool.from_function(example_fn_two))
    assert len(tools_to_openai_spec(tools)) == 2


def test_tools_to_openai_spec_duplicate() -> None:
    tools = (Tool.from_function(example_fn), Tool.from_function(example_fn))
    with pytest.raises(ValueError):
        tools_to_openai_spec(tools)


def test_tools_to_openai_spec_single() -> None:
    tools = Tool.from_function(example_fn)
    expected_spec = [
        {
            "type": "function",
            "function": {
                "name": "example_fn",
                "description": "Testing function.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "param": {
                            "type": "string",
                            "description": "Description of param1",
                        }
                    },
                    "required": ["param"],
                },
            },
        },
    ]
    result = tools_to_openai_spec((tools,))
    assert result == expected_spec


def test_tools_to_openai_spec_empty() -> None:
    tools = ()
    expected_spec = []
    assert tools_to_openai_spec(tools) == expected_spec


def test_openai_response_to_message_text() -> None:
    response = {
        "choices": [
            {
                "role": "assistant",
                "message": {"content": "Hello from John Cena!"},
            }
        ],
        "usage": {
            "input_tokens": 10,
            "output_tokens": 25,
            "total_tokens": 35,
        },
    }

    message = openai_response_to_message(response)

    actual = asdict(message)
    expect = asdict(
        Message(
            role="assistant",
            content=[Text("Hello from John Cena!")],
        )
    )
    actual.pop("id")
    expect.pop("id")
    assert actual == expect


def test_openai_response_to_message_valid_tooluse() -> None:
    response = deepcopy(OPEN_AI_TOOL_USE_RESPONSE)
    message = openai_response_to_message(response)
    actual = asdict(message)
    expect = asdict(
        Message(
            role="assistant",
            content=[ToolUse(id=1, name="example_fn", parameters={"param": "value"})],
        )
    )
    actual.pop("id")
    actual["content"][0].pop("id")
    expect.pop("id")
    expect["content"][0].pop("id")
    assert actual == expect


def test_openai_response_to_message_invalid_func_name() -> None:
    response = deepcopy(OPEN_AI_TOOL_USE_RESPONSE)
    response["choices"][0]["message"]["tool_calls"][0]["function"]["name"] = "invalid fn"
    message = openai_response_to_message(response)
    assert message.content[0].name == "invalid fn"
    assert json.loads(message.content[0].parameters) == {"param": "value"}
    assert message.content[0].is_error
    assert message.content[0].error_message.startswith("The provided function name")


@patch("json.loads", side_effect=json.JSONDecodeError("error", "doc", 0))
def test_openai_response_to_message_json_decode_error(mock_json) -> None:
    response = deepcopy(OPEN_AI_TOOL_USE_RESPONSE)
    message = openai_response_to_message(response)
    assert message.content[0].name == "example_fn"
    assert message.content[0].is_error
    assert message.content[0].error_message.startswith("Could not interpret tool use")



import pytest
from exchange import Exchange
from exchange.checkpoint import Checkpoint
from exchange.message import Message
from exchange.providers import Provider, Usage
from exchange.moderators.truncate import ContextTruncate, pop_checkpoint
from typing import List
from exchange.content import Text
from exchange.tool import Tool


class MockProvider(Provider):
    def __init__(self, sequence: List[Message], usage_dicts: List[dict]):
        # We'll use init to provide a preplanned reply sequence
        self.total_counts = len(sequence)
        self.sequence = sequence
        self.call_count = 0
        self.prev_count = 0
        self.usage_dicts = usage_dicts

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.get("usage", {})
        input_tokens = usage.get("input_tokens")
        output_tokens = usage.get("output_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(self, model: str, system: str, messages: List[Message], tools: List[Tool]) -> Message:
        # truncate moderator does a second call to generate in the first pass, hence this skip
        output = self.sequence[self.call_count % self.total_counts]
        usage = self.get_usage(self.usage_dicts[self.call_count % self.total_counts])
        # truncate moderator does a second call to generate in the first pass, hence this skip
        if self.call_count == 0 and self.prev_count == 0:
            self.prev_count += 1
        else:
            self.call_count += 1
        return (output, usage)


@pytest.fixture
def exchange():
    messages = [Message(role="user", content=[Text(text="Hello")])]
    checkpoints = [Checkpoint(start_index=0, end_index=1, token_count=100, latest_generated_tokens=50)]
    provider = MockProvider(
        sequence=[
            Message(
                role="assistant",
                content=[Text(text="Mock response")],
            ),
            Message(
                role="assistant",
                content=[Text(text="Mock response 2")],
            ),
            Message(
                role="assistant",
                content=[Text(text="This is a long context response")],
            ),
            Message(
                role="assistant",
                content=[Text(text="Normal Mock response")],
            ),
        ],
        usage_dicts=[
            {"usage": {"input_tokens": 100 + 20, "output_tokens": 124000 // 2}},
            {
                "usage": {
                    "input_tokens": (100 + 20 + 124000 // 2) + 30,
                    "output_tokens": 40,
                }
            },
            {
                "usage": {
                    "input_tokens": (100 + 20 + 30 + 40 + 124000 // 2) + 50,
                    "output_tokens": 126000 // 2,
                }
            },
            #  this is where truncation will get adjusted the input tokens for
            # next batch will be total tokens from last two usage dicts
            # which equals 63000 + 50 + 40 + 30
            {"usage": {"input_tokens": 63220 + 11, "output_tokens": 26}},
        ],
    )
    exchange = Exchange(
        provider=provider,
        model="test_model",
        system="test_system",
        moderator=ContextTruncate(),
        tools=[],
        messages=messages,
        checkpoints=checkpoints,
    )
    return exchange


def test_pop_checkpoint(exchange):
    exchange.reply()  # get assistant response and adds to checkpoint/messages
    exchange.add(Message(role="user", content=[Text(text="Hi2")]))
    exchange.reply()  # get assistant response and adds to checkpoint/messages for 2

    new_exchange = pop_checkpoint(exchange)
    assert len(new_exchange.checkpoints) == 2
    assert len(new_exchange.messages) == 3

    assert new_exchange.checkpoints[0].start_index == 0
    assert new_exchange.checkpoints[0].end_index == 1
    assert new_exchange.checkpoints[0].token_count == 124000 // 2 + 20
    assert new_exchange.checkpoints[1].start_index == 1
    assert new_exchange.checkpoints[1].end_index == 3
    assert new_exchange.checkpoints[1].token_count == 30 + 40


def test_context_truncate_rewrite(exchange):
    # exchange.checkpoints.append(Checkpoint(start_index=1, end_index=2, token_count=112000+250,
    # latest_generated_tokens=50))
    exchange.messages.append(Message(role="user", content=[Text(text="user Hi again")]))
    exchange.reply()
    exchange.messages.append(Message(role="user", content=[Text(text="go AI!")]))
    exchange.reply()
    exchange.messages.append(Message(role="user", content=[Text(text="long context coming")]))
    exchange.reply()
    exchange.messages.append(Message(role="user", content=[Text(text="small_context")]))
    exchange.reply()

    # note we create a new exchange in ContextTruncate to get the system_prompt tokens.
    # In this test, it executes the sequence in the MockProvider above and thus the system prompt is 112 tokens
    # exchange.moderator.rewrite()
    assert sum(cp.token_count for cp in exchange.checkpoints) <= 112000
    assert exchange.checkpoints[0].token_count == 190
    assert exchange.checkpoints[0].latest_generated_tokens == 40



from typing import List, Literal, Optional, Tuple, Type, Union

from exchange import Message
from exchange.checkpoint import Checkpoint
from exchange.moderators import ContextTruncate, Moderator, PassiveModerator

MAX_TOKENS = 112000
SUMMARIZATION_OFFSET = 40000  # Keep a max of this many tokens


def pop_checkpoint(
    exchange: Type["exchange.exchange.Exchange"],  # noqa: F821
    exclude_last: Literal[0, 1] = 0,
    return_messages: bool = False,
) -> Union[Tuple[List[Message], Checkpoint], Type["exchange.exchange.Exchange"]]:  # noqa: F821
    """Pop messages from the front of the list in sections

    Inputs:
        exchange (Exchange): An Exchange instance to pop checkpoints off of
        exclude_last (Literal[0,1]): integer flag whether to remove all messages in
            the checkpoint (0), or remove all except the last message (1).
    """
    removed_messages = []
    checkpoint = exchange.checkpoints.pop(0)
    for _ in range(checkpoint.end_index - checkpoint.start_index - exclude_last):
        removed_messages.append(exchange.messages.pop(checkpoint.start_index))

    # rewrite checkpoint indexes
    for cp in exchange.checkpoints:
        cp.start_index = cp.start_index - checkpoint.end_index + exclude_last
        cp.end_index = cp.end_index - checkpoint.end_index + exclude_last

    if exclude_last:
        # there is still one entry in the first checkpoint only
        checkpoint.token_count = checkpoint.latest_generated_tokens
        checkpoint.end_index = exclude_last
        exchange.checkpoints.insert(0, checkpoint)

    if return_messages:
        return removed_messages, checkpoint
    else:
        return exchange


class ContextSummarizer(Moderator):
    def __init__(
        self,
        model: Optional[str] = "gpt-4o-mini",
        max_tokens: Optional[int] = MAX_TOKENS,
        summarization_offset: Optional[int] = SUMMARIZATION_OFFSET,
    ) -> None:
        self.model = model
        self.system_prompt_token_count = None
        self.max_tokens = max_tokens
        self.summarization_offset = summarization_offset

    def rewrite(self, exchange: Type["exchange.exchange.Exchange"]) -> None:  # noqa: F821
        """Summarize the context history up to the last few messages in the exchange"""
        if not self.system_prompt_token_count:
            # calculate the system prompt tokens (includes functions etc...)
            _system_token_exchange = exchange.replace(
                messages=[],
                checkpoints=[],
                moderator=PassiveModerator(),
                model=self.model if self.model else exchange.model,
            )
            _ = _system_token_exchange.generate()
            checkpoint = _system_token_exchange.checkpoints.pop()
            self.system_prompt_token_count = checkpoint.token_count - checkpoint.latest_generated_tokens

        if sum(cp.token_count for cp in exchange.checkpoints) > self.max_tokens:
            # this keeps all the messages/checkpoints
            throwaway_exchange = exchange.replace(
                moderator=PassiveModerator(),
            )

            # keep latest summarization_offset tokens in context, summarize the rest
            removed_checkpoints = 0
            messages_to_summarize = []
            while sum(cp.token_count for cp in throwaway_exchange.checkpoints) > self.summarization_offset:
                m, c = pop_checkpoint(throwaway_exchange, return_messages=True)
                messages_to_summarize.extend(m)
                removed_checkpoints += 1

            exclude_last = False
            if throwaway_exchange.messages[0].tool_result:
                m, c = pop_checkpoint(throwaway_exchange, exclude_last=1, return_messages=True)
                messages_to_summarize.extend(m)
                exclude_last = True

            if messages_to_summarize[-1].role == "assistant" and (not messages_to_summarize[-1].tool_use):
                messages_to_summarize.append(Message.user("Summarize our the above conversation"))

            summarizer_exchange = exchange.replace(
                system=Message.load("summarizer.jinja").text,
                moderator=ContextTruncate(),
                model=self.model,
                messages=messages_to_summarize,
                checkpoints=[],
            )

            summary = summarizer_exchange.reply()
            summary_checkpoint = summarizer_exchange.checkpoints[-1]

            # pop out all the messages up
            for _ in range(removed_checkpoints):
                pop_checkpoint(exchange)

            # if first message is a tool_result, it means there's no matching pair. remove it
            if exclude_last:
                pop_checkpoint(exchange, exclude_last=1)

            # insert summary as first message/checkpoint
            if exchange.messages[0].role == "assistant":
                summary_message = Message.user(summary.text)
            else:
                summary_message = Message.assistant(summary.text)

            exchange.messages.insert(0, summary_message)

            # insert a new checkpoint with the summary message
            exchange.checkpoints.insert(
                0,
                Checkpoint(
                    start_index=0,
                    end_index=1,
                    token_count=summary_checkpoint.latest_generated_tokens + self.system_prompt_token_count,
                    latest_generated_tokens=summary_checkpoint.latest_generated_tokens,
                ),
            )

            # update checkpoint indices
            for cp in exchange.checkpoints[1:]:
                cp.start_index = cp.start_index + 1
                cp.end_index = cp.end_index + 1



import inspect
import time
from pathlib import Path
from typing import Any, Dict, List, Literal, Type

from attrs import define, field
from jinja2 import Environment, FileSystemLoader

from exchange.content import CONTENT_TYPES, Content, Text, ToolResult, ToolUse
from exchange.utils import create_object_id

Role = Literal["user", "assistant"]


def validate_role_and_content(instance: "Message", *_: Any) -> None:  # noqa: ANN401
    if instance.role == "user":
        if not (instance.text or instance.tool_result):
            raise ValueError("User message must include a Text or ToolResult")
        if instance.tool_use:
            raise ValueError("User message does not support ToolUse")
    elif instance.role == "assistant":
        if not (instance.text or instance.tool_use):
            raise ValueError("Assistant message must include a Text or ToolUsage")
        if instance.tool_result:
            raise ValueError("Assistant message does not support ToolResult")


def content_converter(contents: List[Dict[str, Any]]) -> List[Content]:
    return [(CONTENT_TYPES[c.pop("type")](**c) if c.__class__ not in CONTENT_TYPES.values() else c) for c in contents]


@define
class Message:
    """A message to or from a language model.

    This supports several content types to extend to tool usage and (tbi) images.

    We also provide shortcuts for simplified text usage; these two are identical:
    ```
    m = Message(role='user', content=[Text(text='abcd')])
    assert m.content[0].text == 'abcd'

    m = Message.user('abcd')
    assert m.text == 'abcd'
    ```
    """

    role: Role = field(default="user")
    id: str = field(factory=lambda: str(create_object_id(prefix="msg")))
    created: int = field(factory=lambda: int(time.time()))
    content: List[Content] = field(factory=list, validator=validate_role_and_content, converter=content_converter)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "role": self.role,
            "id": self.id,
            "created": self.created,
            "content": [item.to_dict() for item in self.content],
        }

    @property
    def text(self) -> str:
        """The text content of this message."""
        result = []
        for content in self.content:
            if isinstance(content, Text):
                result.append(content.text)
        return "\n".join(result)

    @property
    def tool_use(self) -> List[ToolUse]:
        """All tool use content of this message."""
        result = []
        for content in self.content:
            if isinstance(content, ToolUse):
                result.append(content)
        return result

    @property
    def tool_result(self) -> List[ToolResult]:
        """All tool result content of this message."""
        result = []
        for content in self.content:
            if isinstance(content, ToolResult):
                result.append(content)
        return result

    @classmethod
    def load(
        cls: Type["Message"],
        filename: str,
        role: Role = "user",
        **kwargs: Dict[str, Any],
    ) -> "Message":
        """Load the message from filename relative to where the load is called.

        This only supports simplified content, with a single text entry

        This is meant to emulate importing code rather than a runtime filesystem. So
        if you have a directory of code that contains example.py, and example.py has
        a function that calls User.load('example.jinja'), it will look in the same
        directory as example.py for the jinja file.
        """
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])

        base_path = Path(mod.__file__).parent

        env = Environment(loader=FileSystemLoader(base_path))
        template = env.get_template(filename)
        rendered_content = template.render(**kwargs)

        return cls(role=role, content=[Text(text=rendered_content)])

    @classmethod
    def user(cls: Type["Message"], text: str) -> "Message":
        return cls(role="user", content=[Text(text)])

    @classmethod
    def assistant(cls: Type["Message"], text: str) -> "Message":
        return cls(role="assistant", content=[Text(text)])



import pytest
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.providers import get_provider
from exchange.tool import Tool

too_long_chars = "x" * (2**20 + 1)

cases = [
    (get_provider("openai"), "gpt-4o-mini"),
    (get_provider("databricks"), "databricks-meta-llama-3-70b-instruct"),
    (get_provider("bedrock"), "anthropic.claude-3-5-sonnet-20240620-v1:0"),
]


@pytest.mark.integration  # skipped in CI/CD
@pytest.mark.parametrize("provider,model", cases)
def test_simple(provider, model):
    provider = provider.from_env()

    ex = Exchange(
        provider=provider,
        model=model,
        system="You are a helpful assistant.",
    )

    ex.add(Message.user("Who is the most famous wizard from the lord of the rings"))

    response = ex.reply()

    # It's possible this can be flakey, but in experience so far haven't seen it
    assert "gandalf" in response.text.lower()


@pytest.mark.integration  # skipped in CI/CD
@pytest.mark.parametrize("provider,model", cases)
def test_tools(provider, model):
    provider = provider.from_env()

    def get_password() -> str:
        """Return the password for authentication"""
        return "mellon"

    ex = Exchange(
        provider=provider,
        model=model,
        system="You are a helpful assistant. Expect to need to authenticate using get_password.",
        tools=(Tool.from_function(get_password),),
    )

    ex.add(Message.user("Can you authenticate this session by responding with the password"))

    response = ex.reply()

    # It's possible this can be flakey, but in experience so far haven't seen it
    assert "mellon" in response.text.lower()


@pytest.mark.integration
@pytest.mark.parametrize("provider,model", cases)
def test_tool_use_output_chars(provider, model):
    provider = provider.from_env()

    def get_password() -> str:
        """Return the password for authentication"""
        return too_long_chars

    ex = Exchange(
        provider=provider,
        model=model,
        system="You are a helpful assistant. Expect to need to authenticate using get_password.",
        tools=(Tool.from_function(get_password),),
    )

    ex.add(Message.user("Can you authenticate this session by responding with the password"))

    ex.reply()

    # Without our error handling, this would raise
    # string too long. Expected a string with maximum length 1048576, but got a string with length ...



from attrs import define, field


@define
class Checkpoint:
    """Checkpoint that counts the tokens in messages between the start and end index"""

    start_index: int = field(default=-1)
    end_index: int = field(default=0)
    token_count: int = field(default=0)
    latest_generated_tokens: int = field(default=0)



import os
from unittest.mock import patch

import httpx
import pytest
from exchange import Message, Text
from exchange.content import ToolResult, ToolUse
from exchange.providers.anthropic import AnthropicProvider
from exchange.tool import Tool


def example_fn(param: str) -> None:
    """
    Testing function.

    Args:
        param (str): Description of param1
    """
    pass


@pytest.fixture
@patch.dict(os.environ, {"ANTHROPIC_API_KEY": "test_api_key"})
def anthropic_provider():
    return AnthropicProvider.from_env()


def test_anthropic_response_to_text_message() -> None:
    response = {
        "content": [{"type": "text", "text": "Hello from Claude!"}],
    }
    message = AnthropicProvider.anthropic_response_to_message(response)
    assert message.content[0].text == "Hello from Claude!"


def test_anthropic_response_to_tool_use_message() -> None:
    response = {
        "content": [
            {
                "type": "tool_use",
                "id": "1",
                "name": "example_fn",
                "input": {"param": "value"},
            }
        ],
    }
    message = AnthropicProvider.anthropic_response_to_message(response)
    assert message.content[0].id == "1"
    assert message.content[0].name == "example_fn"
    assert message.content[0].parameters == {"param": "value"}


def test_tools_to_anthropic_spec() -> None:
    tools = (Tool.from_function(example_fn),)
    expected_spec = [
        {
            "name": "example_fn",
            "description": "Testing function.",
            "input_schema": {
                "type": "object",
                "properties": {"param": {"type": "string", "description": "Description of param1"}},
                "required": ["param"],
            },
        }
    ]
    result = AnthropicProvider.tools_to_anthropic_spec(tools)
    assert result == expected_spec


def test_message_text_to_anthropic_spec() -> None:
    messages = [Message.user("Hello, Claude")]
    expected_spec = [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Hello, Claude"}],
        }
    ]
    result = AnthropicProvider.messages_to_anthropic_spec(messages)
    assert result == expected_spec


def test_messages_to_anthropic_spec() -> None:
    messages = [
        Message(role="user", content=[Text(text="Hello, Claude")]),
        Message(
            role="assistant",
            content=[ToolUse(id="1", name="example_fn", parameters={"param": "value"})],
        ),
        Message(role="user", content=[ToolResult(tool_use_id="1", output="Result")]),
    ]
    actual_spec = AnthropicProvider.messages_to_anthropic_spec(messages)
    #  !=
    expected_spec = [
        {"role": "user", "content": [{"type": "text", "text": "Hello, Claude"}]},
        {
            "role": "assistant",
            "content": [
                {
                    "type": "tool_use",
                    "id": "1",
                    "name": "example_fn",
                    "input": {"param": "value"},
                }
            ],
        },
        {
            "role": "user",
            "content": [{"type": "tool_result", "tool_use_id": "1", "content": "Result"}],
        },
    ]
    assert actual_spec == expected_spec


@patch("httpx.Client.post")
@patch("time.sleep", return_value=None)
@patch("logging.warning")
@patch("logging.error")
def test_anthropic_completion(mock_error, mock_warning, mock_sleep, mock_post, anthropic_provider):
    mock_response = {
        "content": [{"type": "text", "text": "Hello from Claude!"}],
        "usage": {"input_tokens": 10, "output_tokens": 25},
    }

    # First 4 attempts fail with status code 429, 5th succeeds
    def create_response(status_code, json_data=None):
        response = httpx.Response(status_code)
        response._content = httpx._content.json_dumps(json_data or {}).encode()
        response._request = httpx.Request("POST", "https://api.anthropic.com/v1/messages")
        return response

    mock_post.side_effect = [
        create_response(429),
        create_response(429),
        create_response(429),
        create_response(429),
        create_response(200, mock_response),
    ]

    model = "claude-3-5-sonnet-20240620"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello, Claude")]

    reply_message, reply_usage = anthropic_provider.complete(model=model, system=system, messages=messages)

    assert reply_message.content == [Text(text="Hello from Claude!")]
    assert reply_usage.total_tokens == 35
    assert mock_post.call_count == 5
    mock_post.assert_any_call(
        "https://api.anthropic.com/v1/messages",
        json={
            "system": system,
            "model": model,
            "max_tokens": 4096,
            "messages": [
                *[
                    {
                        "role": msg.role,
                        "content": [{"type": "text", "text": msg.content[0].text}],
                    }
                    for msg in messages
                ],
            ],
        },
    )


@pytest.mark.integration
def test_anthropic_integration():
    provider = AnthropicProvider.from_env()
    model = "claude-3-5-sonnet-20240620"  # updated model to a known valid model
    system = "You are a helpful assistant."
    messages = [Message.user("Hello, Claude")]

    # Run the completion
    reply = provider.complete(model=model, system=system, messages=messages)

    assert reply[0].content is not None
    print("Completion content from Anthropics:", reply[0].content)



import os
from typing import Any, Dict, List, Tuple, Type

import httpx

from exchange.message import Message
from exchange.providers.base import Provider, Usage
from exchange.providers.utils import (
    messages_to_openai_spec,
    openai_response_to_message,
    openai_single_message_context_length_exceeded,
    raise_for_status,
    tools_to_openai_spec,
)
from exchange.tool import Tool

OPENAI_HOST = "https://api.openai.com/"


class OpenAiProvider(Provider):
    """Provides chat completions for models hosted directly by OpenAI"""

    def __init__(self, client: httpx.Client) -> None:
        super().__init__()
        self.client = client

    @classmethod
    def from_env(cls: Type["OpenAiProvider"]) -> "OpenAiProvider":
        url = os.environ.get("OPENAI_HOST", OPENAI_HOST)
        try:
            key = os.environ["OPENAI_API_KEY"]
        except KeyError:
            raise RuntimeError(
                "Failed to get OPENAI_API_KEY from the environment, see https://platform.openai.com/docs/api-reference/api-keys"
            )
        client = httpx.Client(
            base_url=url,
            auth=("Bearer", key),
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client)

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.pop("usage")
        input_tokens = usage.get("prompt_tokens")
        output_tokens = usage.get("completion_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: Tuple[Tool],
        **kwargs: Dict[str, Any],
    ) -> Tuple[Message, Usage]:
        payload = dict(
            messages=[
                {"role": "system", "content": system},
                *messages_to_openai_spec(messages),
            ],
            model=model,
            tools=tools_to_openai_spec(tools) if tools else [],
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}
        response = self.client.post("v1/chat/completions", json=payload)

        # Check for context_length_exceeded error for single, long input message
        if "error" in response.json() and len(messages) == 1:
            openai_single_message_context_length_exceeded(response.json()["error"])

        data = raise_for_status(response).json()

        message = openai_response_to_message(data)
        usage = self.get_usage(data)
        return message, usage



import pytest
from attr.exceptions import FrozenInstanceError
from exchange.content import Text
from exchange.exchange import Exchange
from exchange.moderators import PassiveModerator
from exchange.message import Message
from exchange.providers import Provider, Usage
from exchange.tool import Tool


def dummy_tool() -> str:
    """An example tool"""
    return "dummy response"


class MockProvider(Provider):
    def complete(self, model, system, messages, tools=None):
        return Message(role="assistant", content=[Text(text="This is a mock response.")]), Usage.from_dict(
            {"total_tokens": 35}
        )


def test_exchange_immutable():
    # Create an instance of Exchange
    provider = MockProvider()
    # intentionally setting a list instead of tuple on tools, it should be converted
    exchange = Exchange(
        provider=provider,
        model="test-model",
        system="test-system",
        tools=(Tool.from_function(dummy_tool),),
        messages=[Message(role="user", content=[Text(text="Hello!")])],
        moderator=PassiveModerator(),
    )

    # Try to directly modify a field (should raise an error)
    with pytest.raises(FrozenInstanceError):
        exchange.system = ""

    with pytest.raises(AttributeError):
        exchange.tools.append("anything")

    # Replace method should return a new instance with deepcopy of messages
    new_exchange = exchange.replace(system="changed")

    assert new_exchange.system == "changed"
    assert len(exchange.messages) == 1
    assert len(new_exchange.messages) == 1

    # Ensure that the messages are deep copied
    new_exchange.messages[0].content[0].text = "Changed!"
    assert exchange.messages[0].content[0].text == "Hello!"
    assert new_exchange.messages[0].content[0].text == "Changed!"



import subprocess
from pathlib import Path
import pytest

from exchange.message import Message
from exchange.content import Text, ToolUse, ToolResult


def test_user_message():
    user_message = Message.user("abcd")
    assert user_message.role == "user"
    assert user_message.text == "abcd"


def test_assistant_message():
    assistant_message = Message.assistant("abcd")
    assert assistant_message.role == "assistant"
    assert assistant_message.text == "abcd"


def test_message_tool_use():
    from exchange.content import ToolUse

    tu1 = ToolUse(id="1", name="tool", parameters={})
    tu2 = ToolUse(id="2", name="tool", parameters={})
    message = Message(role="assistant", content=[tu1, tu2])
    assert len(message.tool_use) == 2
    assert message.tool_use[0].name == "tool"


def test_message_tool_result():
    from exchange.content import ToolResult

    tr1 = ToolResult(tool_use_id="1", output="result")
    tr2 = ToolResult(tool_use_id="2", output="result")
    message = Message(role="user", content=[tr1, tr2])
    assert len(message.tool_result) == 2
    assert message.tool_result[0].output == "result"


def test_message_load(tmpdir):
    # To emulate the expected relative lookup, we need to create a mock code dir
    # and run the load in a subprocess
    test_dir = Path(tmpdir)

    # Create a temporary Jinja template file in the test_dir
    template_content = "hello {{ name }} {% include 'relative.jinja' %}"
    template_path = test_dir / "template.jinja"
    template_path.write_text(template_content)

    relative_content = "and {{ name2 }}"
    relative_path = test_dir / "relative.jinja"
    relative_path.write_text(relative_content)

    # Create a temporary Python file in the sub_dir that calls the load method with a relative path
    python_file_content = """
from exchange.message import Message

def test_function():
    message = Message.load('template.jinja', name="a", name2="b")
    assert message.text == "hello a and b"
    assert message.role == "user"

test_function()
"""
    python_file_path = test_dir / "test_script.py"
    python_file_path.write_text(python_file_content)

    # Execute the temporary Python file to test the relative lookup functionality
    result = subprocess.run(["python3", str(python_file_path)])

    assert result.returncode == 0


def test_message_validation():
    # Valid user message
    message = Message(role="user", content=[Text(text="Hello")])
    assert message.text == "Hello"

    # Valid assistant message
    message = Message(role="assistant", content=[Text(text="Hello")])
    assert message.text == "Hello"

    # Invalid message: user with tool_use
    with pytest.raises(ValueError):
        Message(
            role="user",
            content=[Text(text=""), ToolUse(id="1", name="tool", parameters={})],
        )

    # Invalid message: assistant with tool_result
    with pytest.raises(ValueError):
        Message(
            role="assistant",
            content=[Text(text=""), ToolResult(tool_use_id="1", output="result")],
        )



from functools import cache
from typing import Type

from exchange.moderators.base import Moderator
from exchange.utils import load_plugins
from exchange.moderators.passive import PassiveModerator  # noqa
from exchange.moderators.truncate import ContextTruncate  # noqa
from exchange.moderators.summarizer import ContextSummarizer  # noqa


@cache
def get_moderator(name: str) -> Type[Moderator]:
    return load_plugins(group="exchange.moderator")[name]



import hashlib
import hmac
import json
import logging
import os
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Tuple, Type
from urllib.parse import quote, urlparse

import httpx

from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.providers import Provider, Usage
from exchange.providers.utils import raise_for_status
from exchange.tool import Tool

SERVICE = "bedrock-runtime"
UTC = timezone.utc

logger = logging.getLogger(__name__)


class AwsClient(httpx.Client):
    def __init__(
        self,
        aws_region: str,
        aws_access_key: str,
        aws_secret_key: str,
        aws_session_token: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.region = aws_region
        self.host = f"https://{SERVICE}.{aws_region}.amazonaws.com/"
        self.access_key = aws_access_key
        self.secret_key = aws_secret_key
        self.session_token = aws_session_token
        super().__init__(base_url=self.host, **kwargs)

    def post(self, path: str, json: Dict, **kwargs: Dict[str, Any]) -> httpx.Response:
        signed_headers = self.sign_and_get_headers(
            method="POST",
            url=path,
            payload=json,
            service="bedrock",
        )
        return super().post(url=path, headers=signed_headers, **kwargs)

    def sign_and_get_headers(
        self,
        method: str,
        url: str,
        payload: dict,
        service: str,
    ) -> Dict[str, str]:
        """
        Sign the request and generate the necessary headers for AWS authentication.

        Args:
            method (str): HTTP method (e.g., 'GET', 'POST').
            url (str): The request URL.
            payload (dict): The request payload.
            service (str): The AWS service name.
            region (str): The AWS region.
            access_key (str): The AWS access key.
            secret_key (str): The AWS secret key.
            session_token (Optional[str]): The AWS session token, if any.

        Returns:
            Dict[str, str]: The headers required for the request.
        """

        def sign(key: bytes, msg: str) -> bytes:
            return hmac.new(key, msg.encode("utf-8"), hashlib.sha256).digest()

        def get_signature_key(key: str, date_stamp: str, region_name: str, service_name: str) -> bytes:
            k_date = sign(("AWS4" + key).encode("utf-8"), date_stamp)
            k_region = sign(k_date, region_name)
            k_service = sign(k_region, service_name)
            k_signing = sign(k_service, "aws4_request")
            return k_signing

        # Convert payload to JSON string
        request_parameters = json.dumps(payload)

        # Create a date for headers and the credential string
        t = datetime.now(UTC)
        amz_date = t.strftime("%Y%m%dT%H%M%SZ")
        date_stamp = t.strftime("%Y%m%d")  # Date w/o time, used in credential scope

        # Create canonical URI and headers
        parsedurl = urlparse(url)
        canonical_uri = quote(parsedurl.path if parsedurl.path else "/", safe="/-_.~")
        canonical_headers = f"host:{parsedurl.netloc}\nx-amz-date:{amz_date}\n"

        # Create the list of signed headers.
        signed_headers = "host;x-amz-date"
        if self.session_token:
            canonical_headers += "x-amz-security-token:" + self.session_token + "\n"
            signed_headers += ";x-amz-security-token"

        # Create payload hash
        payload_hash = hashlib.sha256(request_parameters.encode("utf-8")).hexdigest()

        # Canonical request
        canonical_request = f"{method}\n{canonical_uri}\n\n{canonical_headers}\n{signed_headers}\n{payload_hash}"

        # Create the string to sign
        algorithm = "AWS4-HMAC-SHA256"
        credential_scope = f"{date_stamp}/{self.region}/{service}/aws4_request"
        string_to_sign = (
            f'{algorithm}\n{amz_date}\n{credential_scope}\n'
            f'{hashlib.sha256(canonical_request.encode("utf-8")).hexdigest()}'
        )

        # Create the signing key
        signing_key = get_signature_key(self.secret_key, date_stamp, self.region, service)

        # Sign the string_to_sign using the signing key
        signature = hmac.new(signing_key, string_to_sign.encode("utf-8"), hashlib.sha256).hexdigest()

        # Add signing information to the request
        authorization_header = (
            f"{algorithm} Credential={self.access_key}/{credential_scope}, SignedHeaders={signed_headers},"
            f"Signature={signature}"
        )

        # Headers
        headers = {
            "Content-Type": "application/json",
            "Authorization": authorization_header,
            "X-Amz-date": amz_date.encode(),
            "x-amz-content-sha256": payload_hash,
        }
        if self.session_token:
            headers["X-Amz-Security-Token"] = self.session_token

        return headers


class BedrockProvider(Provider):
    def __init__(self, client: AwsClient) -> None:
        self.client = client

    @classmethod
    def from_env(cls: Type["BedrockProvider"]) -> "BedrockProvider":
        aws_region = os.environ.get("AWS_REGION", "us-east-1")
        try:
            aws_access_key = os.environ["AWS_ACCESS_KEY_ID"]
            aws_secret_key = os.environ["AWS_SECRET_ACCESS_KEY"]
            aws_session_token = os.environ.get("AWS_SESSION_TOKEN")
        except KeyError:
            raise RuntimeError("Failed to get AWS_ACCESS_KEY_ID or AWS_SECRET_ACCESS_KEY from the environment")

        client = AwsClient(
            aws_region=aws_region,
            aws_access_key=aws_access_key,
            aws_secret_key=aws_secret_key,
            aws_session_token=aws_session_token,
        )
        return cls(client=client)

    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: Tuple[Tool],
        **kwargs: Dict[str, Any],
    ) -> Tuple[Message, Usage]:
        """
        Generate a completion response from the Bedrock gateway.

        Args:
            model (str): The model identifier.
            system (str): The system prompt or configuration.
            messages (List[Message]): A list of messages to be processed by the model.
            tools (Tuple[Tool]): A tuple of tools to be used in the completion process.
            **kwargs: Additional keyword arguments for inference configuration.

        Returns:
            Tuple[Message, Usage]: A tuple containing the response message and usage data.
        """

        inference_config = dict(
            temperature=kwargs.pop("temperature", None),
            maxTokens=kwargs.pop("max_tokens", None),
            stopSequences=kwargs.pop("stop", None),
            topP=kwargs.pop("topP", None),
        )
        inference_config = {k: v for k, v in inference_config.items() if v is not None} or None

        converted_messages = [self.message_to_bedrock_spec(message) for message in messages]
        converted_system = [dict(text=system)]
        tool_config = self.tools_to_bedrock_spec(tools)
        payload = dict(
            system=converted_system,
            inferenceConfig=inference_config,
            messages=converted_messages,
            toolConfig=tool_config,
            **kwargs,
        )
        payload = {k: v for k, v in payload.items() if v}

        path = f"model/{model}/converse"

        response = self.client.post(path, json=payload)
        raise_for_status(response)
        response_message = response.json()["output"]["message"]

        usage_data = response.json()["usage"]
        usage = Usage(
            input_tokens=usage_data.get("inputTokens"),
            output_tokens=usage_data.get("outputTokens"),
            total_tokens=usage_data.get("totalTokens"),
        )

        return self.response_to_message(response_message), usage

    @staticmethod
    def message_to_bedrock_spec(message: Message) -> dict:
        bedrock_content = []
        try:
            for content in message.content:
                if isinstance(content, Text):
                    bedrock_content.append({"text": content.text})
                elif isinstance(content, ToolUse):
                    for tool_use in message.tool_use:
                        bedrock_content.append(
                            {
                                "toolUse": {
                                    "toolUseId": tool_use.id,
                                    "name": tool_use.name,
                                    "input": tool_use.parameters,
                                }
                            }
                        )
                elif isinstance(content, ToolResult):
                    for tool_result in message.tool_result:
                        # try to parse the output as json
                        try:
                            output = json.loads(tool_result.output)
                            if isinstance(output, dict):
                                content = [{"json": output}]
                            else:
                                content = [{"text": str(output)}]
                        except json.JSONDecodeError:
                            content = [{"text": tool_result.output}]

                        bedrock_content.append(
                            {
                                "toolResult": {
                                    "toolUseId": tool_result.tool_use_id,
                                    "content": content,
                                    **({"status": "error"} if tool_result.is_error else {}),
                                }
                            }
                        )
            return {"role": message.role, "content": bedrock_content}

        except AttributeError:
            raise Exception("Invalid message")

    @staticmethod
    def response_to_message(response_message: dict) -> Message:
        content = []
        if response_message["role"] == "user":
            for block in response_message["content"]:
                if "text" in block:
                    content.append(Text(block["text"]))
                if "toolResult" in block:
                    content.append(
                        ToolResult(
                            tool_use_id=block["toolResult"]["toolResultId"],
                            output=block["toolResult"]["content"][0]["json"],
                            is_error=block["toolResult"].get("status") == "error",
                        )
                    )
            return Message(role="user", content=content)
        elif response_message["role"] == "assistant":
            for block in response_message["content"]:
                if "text" in block:
                    content.append(Text(block["text"]))
                if "toolUse" in block:
                    content.append(
                        ToolUse(
                            id=block["toolUse"]["toolUseId"],
                            name=block["toolUse"]["name"],
                            parameters=block["toolUse"]["input"],
                        )
                    )
            return Message(role="assistant", content=content)
        raise Exception("Invalid response")

    @staticmethod
    def tools_to_bedrock_spec(tools: Tuple[Tool]) -> Optional[dict]:
        if len(tools) == 0:
            return None  # API requires a non-empty tool config or None
        tools_added = set()
        tool_config_list = []
        for tool in tools:
            if tool.name in tools_added:
                logging.warning(f"Tool {tool.name} already added to tool config. Skipping.")
                continue
            tool_config_list.append(
                {
                    "toolSpec": {
                        "name": tool.name,
                        "description": tool.description,
                        "inputSchema": {"json": tool.parameters},
                    }
                }
            )
            tools_added.add(tool.name)
        tool_config = {"tools": tool_config_list}
        return tool_config



from typing import List

from exchange.content import Text, ToolResult, ToolUse
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.moderators import PassiveModerator
from exchange.providers import Provider, Usage
from exchange.tool import Tool


def dummy_tool() -> str:
    """An example tool"""
    return "dummy response"


too_long_output = "x" * (2**20 + 1)
too_long_token_output = "word " * 128000


class MockProvider(Provider):
    def __init__(self, sequence: List[Message], usage_dicts: List[dict]):
        # We'll use init to provide a preplanned reply sequence
        self.sequence = sequence
        self.call_count = 0
        self.usage_dicts = usage_dicts

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.pop("usage")
        input_tokens = usage.get("input_tokens")
        output_tokens = usage.get("output_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(self, model: str, system: str, messages: List[Message], tools: List[Tool]) -> Message:
        output = self.sequence[self.call_count]
        usage = self.get_usage(self.usage_dicts[self.call_count])
        self.call_count += 1
        return (output, usage)


def test_reply_with_unsupported_tool():
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="unsupported_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=(Tool.from_function(dummy_tool),),
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert isinstance(content, ToolResult) and content.is_error and "no tool exists" in content.output.lower()


def test_invalid_tool_parameters():
    """Test handling of invalid tool parameters response"""
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="dummy_tool", parameters="invalid json")],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test invalid parameters")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert isinstance(content, ToolResult) and content.is_error and "invalid json" in content.output.lower()


def test_max_tool_use_when_limit_reached():
    """Test the max_tool_use parameter in the reply method."""
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="dummy_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[ToolUse(id="2", name="dummy_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[ToolUse(id="3", name="dummy_tool", parameters={})],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test max tool use")]))

    response = ex.reply(max_tool_use=3)

    assert ex.provider.call_count == 3
    assert "reached the limit" in response.content[0].text.lower()

    assert isinstance(ex.messages[-2].content[0], ToolResult) and ex.messages[-2].content[0].tool_use_id == "3"

    assert ex.messages[-1].role == "assistant"


def test_tool_output_too_long_character_error():
    """Test tool handling when output exceeds character limit."""

    def long_output_tool_char() -> str:
        return too_long_output

    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="long_output_tool_char", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(long_output_tool_char)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test long output char")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert (
        isinstance(content, ToolResult)
        and content.is_error
        and "output that was too long to handle" in content.output.lower()
    )


def test_tool_output_too_long_token_error():
    """Test tool handling when output exceeds token limit."""

    def long_output_tool_token() -> str:
        return too_long_token_output

    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="long_output_tool_token", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(long_output_tool_token)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test long output token")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert (
        isinstance(content, ToolResult)
        and content.is_error
        and "output that was too long to handle" in content.output.lower()
    )


def test_usage_param():
    usage_tests = {
        0: ({"usage": {"total_tokens": 35}}, 35),
        1: ({"usage": {"input_tokens": 12, "output_tokens": 23}}, 35),
    }

    for count, (usage_dict, outcome) in usage_tests.items():
        ex = Exchange(
            provider=MockProvider(
                sequence=[
                    Message(
                        role="assistant",
                        content=[Text(text="Here is the completion after tool call")],
                    ),
                ],
                usage_dicts=[usage_dict],
            ),
            model="gpt-4o-2024-05-13",
            system="You are a helpful assistant.",
            tools=(Tool.from_function(dummy_tool),),
            moderator=PassiveModerator(),
        )

        ex.add(Message(role="user", content=[Text(text="test")]))

        ex.reply()
        assert ex.checkpoints.pop(-1).token_count == outcome


def test_checkpoints_on_exchange():
    """Test checkpoints on an exchange."""
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(role="assistant", content=[Text(text="Message 1")]),
                Message(role="assistant", content=[Text(text="Message 2")]),
                Message(role="assistant", content=[Text(text="Message 3")]),
            ],
            usage_dicts=[
                {"usage": {"total_tokens": 10}},
                {"usage": {"total_tokens": 28}},
                {"usage": {"total_tokens": 33}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=(Tool.from_function(dummy_tool),),
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()

    # Check if checkpoints are created correctly
    checkpoints = ex.checkpoints
    print(checkpoints)
    assert len(checkpoints) == 3
    assert checkpoints[0].token_count == 10
    assert checkpoints[1].token_count == 18
    assert checkpoints[2].token_count == 5

    # Check if the messages are ordered correctly
    assert [msg.content[0].text for msg in ex.messages] == [
        "User message",
        "Message 1",
        "User message",
        "Message 2",
        "User message",
        "Message 3",
    ]



import pytest
from exchange import Exchange, Message
from exchange.checkpoint import Checkpoint
from exchange.moderators.summarizer import ContextSummarizer
from exchange.providers import Usage


class MockProvider:
    def complete(self, model, system, messages, tools):
        assistant_message_text = "Summarized content here."
        output_tokens = len(assistant_message_text)
        total_input_tokens = sum(len(msg.text) for msg in messages)
        message = Message.assistant(assistant_message_text)
        total_tokens = total_input_tokens + output_tokens
        usage = Usage(input_tokens=total_input_tokens, output_tokens=output_tokens, total_tokens=total_tokens)
        return message, usage


@pytest.fixture
def exchange_instance():
    return Exchange(
        provider=MockProvider(),
        model="test-model",
        system="test-system",
        messages=[
            Message.user("Hi, can you help me with my homework?"),
            Message.assistant("Of course! What do you need help with?"),
            Message.user("I need help with math problems."),
            Message.assistant("Sure, I can help with that. Let's get started."),
            Message.user("Can you also help with my science homework?"),
            Message.assistant("Yes, I can help with science too."),
            Message.user("That's great! How about history?"),
            Message.assistant("Of course, I can help with history as well."),
            Message.user("Thanks! You're very helpful."),
            Message.assistant("You're welcome! I'm here to help."),
        ],
        checkpoints=[
            Checkpoint(start_index=0, end_index=2, token_count=134, latest_generated_tokens=23),
            Checkpoint(start_index=2, end_index=4, token_count=135, latest_generated_tokens=112),
            Checkpoint(start_index=4, end_index=6, token_count=143, latest_generated_tokens=31),
            Checkpoint(start_index=6, end_index=8, token_count=135, latest_generated_tokens=104),
            Checkpoint(start_index=8, end_index=10, token_count=139, latest_generated_tokens=35),
        ],
    )


@pytest.fixture
def summarizer_instance():
    return ContextSummarizer(max_tokens=500, summarization_offset=200)


def test_context_summarizer_rewrite(exchange_instance, summarizer_instance):
    # Pre-checks
    assert len(exchange_instance.messages) == 10

    # Perform summarization
    summarizer_instance.rewrite(exchange_instance)

    # Post-checks
    total_tokens = sum(cp.token_count for cp in exchange_instance.checkpoints)
    assert total_tokens <= 200

    # Assert that summarized content is the first message
    first_message = exchange_instance.messages[0]
    assert first_message.role == "user" or first_message.role == "assistant"
    assert any("summarized" in content.text.lower() for content in first_message.content)

    # Ensure roles alternate in the output
    for i in range(len(exchange_instance.messages) - 1):
        assert (
            exchange_instance.messages[i].role != exchange_instance.messages[i + 1].role
        ), "Messages must alternate between user and assistant"



import os
from unittest.mock import patch

import pytest
from exchange import Message, Text
from exchange.providers.openai import OpenAiProvider


@pytest.fixture
@patch.dict(os.environ, {"OPENAI_API_KEY": "test_api_key"})
def openai_provider():
    return OpenAiProvider.from_env()


@patch("httpx.Client.post")
@patch("time.sleep", return_value=None)
@patch("logging.warning")
@patch("logging.error")
def test_openai_completion(mock_error, mock_warning, mock_sleep, mock_post, openai_provider):
    mock_response = {
        "choices": [{"message": {"role": "assistant", "content": "Hello!"}}],
        "usage": {"prompt_tokens": 10, "completion_tokens": 25, "total_tokens": 35},
    }

    mock_post.return_value.json.return_value = mock_response

    model = "gpt-4"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    tools = ()

    reply_message, reply_usage = openai_provider.complete(model=model, system=system, messages=messages, tools=tools)

    assert reply_message.content == [Text(text="Hello!")]
    assert reply_usage.total_tokens == 35
    mock_post.assert_called_once_with(
        "v1/chat/completions",
        json={
            "messages": [
                {"role": "system", "content": system},
                {"role": "user", "content": "Hello"},
            ],
            "model": model,
        },
    )


@pytest.mark.integration
def test_openai_integration():
    provider = OpenAiProvider.from_env()
    model = "gpt-4"  # specify a valid model
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]

    reply = provider.complete(model=model, system=system, messages=messages, tools=None)

    assert reply[0].content is not None
    print("Completion content from OpenAI:", reply[0].content)



import inspect
import uuid
from importlib.metadata import entry_points
from typing import Any, Callable, Dict, List, Type, get_args, get_origin

from griffe import (
    Docstring,
    DocstringSection,
    DocstringSectionParameters,
    DocstringSectionText,
)


def create_object_id(prefix: str) -> str:
    return f"{prefix}_{uuid.uuid4().hex[:24]}"


def compact(content: str) -> str:
    """Replace any amount of whitespace with a single space"""
    return " ".join(content.split())


def parse_docstring(func: Callable) -> tuple[str, List[Dict]]:
    """Get description and parameters from function docstring"""
    function_args = list(inspect.signature(func).parameters.keys())
    text = str(func.__doc__)
    docstring = Docstring(text)

    for style in ["google", "numpy", "sphinx"]:
        parsed = docstring.parse(style)

        if not _check_section_is_present(parsed, DocstringSectionText):
            continue

        if function_args and not _check_section_is_present(parsed, DocstringSectionParameters):
            continue
        break
    else:  # if we did not find a valid style in the for loop
        raise ValueError(
            f"Attempted to load from a function {func.__name__} with an invalid docstring. Parameter docs are required if the function has parameters. https://mkdocstrings.github.io/griffe/reference/docstrings/#docstrings"  # noqa: E501
        )

    description = None
    parameters = []

    for section in parsed:
        if isinstance(section, DocstringSectionText):
            description = compact(section.value)
        elif isinstance(section, DocstringSectionParameters):
            parameters = [arg.as_dict() for arg in section.value]

    docstring_args = [d["name"] for d in parameters]
    if description is None:
        raise ValueError("Docstring must include a description.")

    if not docstring_args == function_args:
        extra_docstring_args = ", ".join(sorted(set(docstring_args) - set(function_args)))
        extra_function_args = ", ".join(sorted(set(function_args) - set(docstring_args)))
        if extra_docstring_args and extra_function_args:
            raise ValueError(
                f"Docstring args must match function args: docstring had extra {extra_docstring_args}; function had extra {extra_function_args}"  # noqa: E501
            )
        elif extra_function_args:
            raise ValueError(f"Docstring args must match function args: function had extra {extra_function_args}")
        elif extra_docstring_args:
            raise ValueError(f"Docstring args must match function args: docstring had extra {extra_docstring_args}")
        else:
            raise ValueError("Docstring args must match function args")

    return description, parameters


def _check_section_is_present(
    parsed_docstring: List[DocstringSection], section_type: Type[DocstringSectionText]
) -> bool:
    for section in parsed_docstring:
        if isinstance(section, section_type):
            return True
    return False


def json_schema(func: Any) -> dict[str, Any]:  # noqa: ANN401
    """Get the json schema for a function"""
    signature = inspect.signature(func)
    parameters = signature.parameters

    schema = {
        "type": "object",
        "properties": {},
        "required": [],
    }

    for param_name, param in parameters.items():
        param_schema = {}

        if param.annotation is not inspect.Parameter.empty:
            param_schema = _map_type_to_schema(param.annotation)

        if param.default is not inspect.Parameter.empty:
            param_schema["default"] = param.default

        schema["properties"][param_name] = param_schema

        if param.default is inspect.Parameter.empty:
            schema["required"].append(param_name)

    return schema


def _map_type_to_schema(py_type: Type) -> Dict[str, Any]:  # noqa: ANN401
    origin = get_origin(py_type)
    args = get_args(py_type)

    if origin is list or origin is tuple:
        return {"type": "array", "items": _map_type_to_schema(args[0] if args else Any)}
    elif origin is dict:
        return {
            "type": "object",
            "additionalProperties": _map_type_to_schema(args[1] if len(args) > 1 else Any),
        }
    elif py_type is int:
        return {"type": "integer"}
    elif py_type is bool:
        return {"type": "boolean"}
    elif py_type is float:
        return {"type": "number"}
    elif py_type is str:
        return {"type": "string"}
    else:
        return {"type": "string"}


def load_plugins(group: str) -> dict:
    """
    Load plugins based on a specified entry point group.

    This function iterates through all entry points registered under a specified group

    Args:
        group (str): The entry point group to load plugins from. This should match the group specified
                     in the package setup where plugins are defined.

    Returns:
        dict: A dictionary where each key is the entry point name, and the value is the loaded plugin object.

    Raises:
        Exception: Propagates exceptions raised by entry point loading, which might occur if a plugin
                   is not found or if there are issues with the plugin's code.
    """
    plugins = {}
    # Access all entry points for the specified group and load each.
    for entrypoint in entry_points(group=group):
        plugin = entrypoint.load()  # Load the plugin.
        plugins[entrypoint.name] = plugin  # Store the loaded plugin in the dictionary.
    return plugins



import pytest
from exchange.content import ToolResult, ToolUse
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.providers import get_provider


cases = [
    (get_provider("openai"), "gpt-4o-mini"),
]


@pytest.mark.integration  # skipped in CI/CD
@pytest.mark.parametrize("provider,model", cases)
def test_simple(provider, model):
    provider = provider.from_env()

    ex = Exchange(
        provider=provider,
        model=model,
        system="You are a helpful assistant.",
    )

    ex.add(Message.user("What does the first entry in the menu say?"))
    ex.add(
        Message(
            role="assistant",
            content=[ToolUse(id="xyz", name="screenshot", parameters={})],
        )
    )
    ex.add(
        Message(
            role="user",
            content=[ToolResult(tool_use_id="xyz", output='"image:tests/test_image.png"')],
        )
    )

    response = ex.reply()

    # It's possible this can be flakey, but in experience so far haven't seen it
    assert "ask goose" in response.text.lower()



from abc import ABC, abstractmethod
from attrs import define, field
from typing import List, Tuple, Type

from exchange.message import Message
from exchange.tool import Tool


@define
class Usage:
    input_tokens: int = field(factory=None)
    output_tokens: int = field(default=None)
    total_tokens: int = field(default=None)


class Provider(ABC):
    @classmethod
    def from_env(cls: Type["Provider"]) -> "Provider":
        return cls()

    @abstractmethod
    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: Tuple[Tool],
    ) -> Tuple[Message, Usage]:
        """Generate the next message using the specified model"""
        pass
